# Gridlock - a collection of grid-based strategy games
# Copyright (c) 2009 Brian Nenninger (brian@dozingcatsoftware.com)
# Gridlock comes with ABSOLUTELY NO WARRANTY, and is released under version 3 of the GNU GPL.

import GridUtils
from Game import GameState
from GenericAI import GenericAI
import LineGame

def capture_pieces_from_move(grid, pnum, pos, ncap):
    """ Removes lines of ncap pieces that are captured by the move at pos. Returns total number removed. """
    remove = []
    for d in GridUtils.direction_array(grid.ndim):
        remdir = []
        # check for ncap enemy pieces followed by friendly piece
        for n, (lpos, val) in enumerate(GridUtils.enumerate_values_with_direction(grid, pos, d)):
            if val==0: break
            # last item we check should be friendly, if so add to master remove list
            if n==ncap:
                if val==pnum:
                    remove.extend(remdir)
                break
            else:
                # if enemy add to potential remove list, otherwise abort
                if val==pnum:
                    break
                else:
                    remdir.append(lpos)        
    for rp in remove:
        grid[rp] = 0
    return len(remove)            


class PenteGameState (LineGame.LineGameState): 
    def reset(self):
        GameState.reset(self)
        self.extra_state = {'player_scores': [0,0]}

    def player_score(self, pnum):
        return self.extra_state['player_scores'][pnum-1]
        
    def has_scores(self):
        return True

    def copy_extra_state(self):
        return {'player_scores': self.extra_state['player_scores'][:]}
        
    def winning_score(self):
        return self.game_params.get('win_score', 10)
        
    def capture_lengths(self):
        return self.game_params.get('captures', [2])
        
    def winner_by_score(self):
        for index,score in enumerate(self.extra_state['player_scores']):
            if score>=self.winning_score():
                return index+1
        return 0

    def is_game_over(self):
        if self.winner_by_score()>0:
            return True
        return LineGame.LineGameState.is_game_over(self)
        
    def winning_player(self):
        ws = self.winner_by_score()
        if ws>0:
            return ws
        return LineGame.LineGameState.winning_player(self)

    def successors(self):
        for c in self.enumerate_coords_with_value(0):
            nextgrid = self.grid.copy()
            nextgrid[c] = self.current_player
            score = sum(capture_pieces_from_move(nextgrid, self.current_player, c, caplen) for caplen in self.capture_lengths())
            
            nextstate = self.next_state_with_grid(nextgrid)
            nextstate.extra_state['player_scores'][self.current_player-1] += score
            yield ( (c,), nextstate )
            
    def win_length(self):
        return 5
            
NEXT_PLAYER_WEIGHTS = [0,0, 1,5, 10,50, 100,500, 1000,50000, 9999999,9999999]
CURRENT_PLAYER_WEIGHTS = [0,0, 1,5, 10,50, 500,2500, 999999,999999, 9999999,9999999]
SCORE_WEIGHT = 1000
            
class PenteAI (GenericAI):
    def game_utility(self, gstate):
        utility = LineGame.threat_weight(gstate, gstate.current_player, CURRENT_PLAYER_WEIGHTS) - \
                  LineGame.threat_weight(gstate, gstate.next_player(), NEXT_PLAYER_WEIGHTS)
        scorediff = (gstate.player_score(gstate.current_player) - gstate.player_score(gstate.next_player()))
        return utility + SCORE_WEIGHT * scorediff
